;; Generated package description from rcirc-sqlite.el  -*- no-byte-compile: t -*-
(define-package "rcirc-sqlite" "1.0.4.0.20241108.161537" "rcirc logging in SQLite" '((emacs "30.0")) :commit "349aabcbb6bd10b007f431d180e870a5918b9e21" :authors '(("Matto Fransen" . "matto@matto.nl")) :maintainer '("Matto Fransen" . "matto@matto.nl") :keywords '("comm") :url "https://codeberg.org/mattof/rcirc-sqlite")
